__version__ = "0.1.0"
"""
GeoPlanar

A module for handling planar enforcement for polygon geoseries.
"""

from geoplanar.overlap import trim_overlaps, is_overlapping
from geoplanar.hole import (holes, fill_holes, add_interiors,
                            missing_interiors)
from geoplanar.planar import (planar_enforce, is_planar_enforced,
                              fix_npe_edges, insert_intersections,
                              non_planar_edges)
from geoplanar.valid import isvalid




#!/usr/bin/env python3
import libpysal
import geopandas
from collections import defaultdict
from shapely.geometry import box, LineString, Polygon, Point
from shapely.ops import split, linemerge
from .overlap import is_overlapping
from .hole import holes


def non_planar_edges(gdf):
    """Find coincident nonplanar edges


    Parameters
    ----------

    gdf :  GeoDataFrame with polygon (multipolygon) GeoSeries


    Returns
    -------

    missing : dictionary
              key is origin feature, value is neighboring feature for each pair
              of coincident nonplanar edges

    Examples
    --------
    >>> c1 = [[0,0], [0, 10], [10, 10], [10, 0], [0, 0]]
    >>> p1 = Polygon(c1)
    >>> c2 = [[10, 2], [10, 8], [20, 8], [20, 2], [10, 2]]
    >>> p2 = Polygon(c2)
    >>> gdf = geopandas.GeoDataFrame(geometry=[p1, p2])
    >>> geoplanar.non_planar_edges(gdf)
    defaultdict(set, {0: {1}})

    """
    w = libpysal.weights.Queen.from_dataframe(gdf)
    intersections = gdf.sindex.query_bulk(gdf.geometry, predicate='intersects').T
    w1 = defaultdict(list)
    for i,j in intersections:
        if i != j:
            w1[i].append(j)
    missing = defaultdict(set)
    for key in w.neighbors:
        if set(w.neighbors[key]) != set(w1[key]):
            for value in w1[key]:
                pair = [key, value]
                pair.sort()
                l, r = pair
                missing[l] = missing[l].union([r])
    return missing

def planar_enforce(gdf):
    uu = gdf.unary_union
    geoms = [uu.intersection(row.geometry) for index, row in gdf.iterrows()]
    return geopandas.GeoDataFrame(geometry=geoms)

def is_planar_enforced(gdf):
    """Test if a geodataframe has any planar enforcement violations


    Parameters
    ----------


    Returns
    -------
    boolean
    """

    if is_overlapping(gdf):
        return False
    if non_planar_edges(gdf):
        return False
    _holes = holes(gdf)
    if _holes.shape[0] > 0:
        return False
    return True

def fix_npe_edges(gdf, inplace=False):
    """Fix all npe intersecting edges in geoseries.

    Arguments
    ---------

    gdf: GeoDataFrame with polygon geoseries for geometry


    Returns
    -------
    gdf: GeoDataFrame with geometries respected planar edges.

    Examples
    --------
    >>> c1 = [[0,0], [0, 10], [10, 10], [10, 0], [0, 0]]
    >>> p1 = Polygon(c1)
    >>> c2 = [[10, 2], [10, 8], [20, 8], [20, 2], [10, 2]]
    >>> p2 = Polygon(c2)
    >>> gdf = geopandas.GeoDataFrame(geometry=[p1, p2])
    >>> geoplanar.non_planar_edges(gdf)
    defaultdict(set, {0: {1}})

    >>> gdf1 = geoplanar.fix_npe_edges(gdf)
    >>> geoplanar.non_planar_edges(gdf1)
    defaultdict(set, {})


    """

    if not inplace:
        gdf = gdf.copy()

    edges = non_planar_edges(gdf)
    for key in edges:
        poly_a = gdf.iloc[key].geometry
        for j in edges[key]:
            poly_b = gdf.iloc[j].geometry
            new_a, new_b = insert_intersections(poly_a, poly_b)
            poly_a = new_a
            gdf.geometry[key] = new_a
            gdf.geometry[j] = new_b
    return gdf

def insert_intersections(poly_a, poly_b):
    """Correct two npe intersecting polygons by inserting intersection points on intersecting edges
    """

    pint = poly_a.intersection(poly_b)
    if isinstance(pint, LineString): # npe edges overlap
        new_polys = []
        pnts = pint
        for poly in [poly_a, poly_b]:
            exterior = LineString(list(poly.exterior.coords))
            sp = [Point(pnt) for pnt in list(zip(*pnts.coords.xy))]
            for pnt in sp:
                splits = split(exterior, pnt)
                if len(splits) > 1:
                    left, right = split(exterior, pnt)
                    exterior = linemerge([left, right])
            new_polys.append(Polygon(list(zip(*exterior.coords.xy))))
        return new_polys
    else:
        lin_a = LineString(list(poly_a.exterior.coords))
        lin_b = LineString(list(poly_b.exterior.coords))
        new_polys = []
        pts = lin_a.intersection(lin_b)
        for lin in [lin_a, lin_b]:
            splits = split(lin, pts)
            coords = [list(zip(*line.coords.xy)) for line in list(splits)]
            new_poly = coords[0]
            for coord in coords[1:]:
                if coord[0] == new_poly[-1]:
                    new_poly.extend(coord[1:])
                else:
                    new_poly.extend(coord)
            new_polys.append(Polygon(new_poly))
        return new_polys



#!/usr/bin/env python3
#
import pandas
import geopandas
from shapely.geometry import box


def holes(gdf):
    """Find holes in a geodataframe.

    A hole is also known as a sliver which contains a set of points that:

    - are not contained by any of the geometries in the geoseries
    - are contained in the convex hull of the unary_union of the geoseries

    Parameters
    ----------

    gdf :  GeoDataFrame with polygon (multipolygon) GeoSeries


    Returns
    -------

    _holes : GeoDataFrame with hole polygons

    Examples
    --------
    >>> p1 = box(0,0,10,10)
    >>> p2 = Polygon([(10,10), (12,8), (10,6), (12,4), (10,2), (20,5)])
    >>> gdf = geopandas.GeoDataFrame(geometry=[p1,p2])
    >>> h = geoplanar.holes(gdf)
    >>> h.area
    array([4., 4.])
    """

    # get box
    b = box(*gdf.total_bounds)

    # form unary union
    u = gdf.unary_union

    # diff of b and u
    dbu = geopandas.GeoDataFrame(geometry=[b.difference(u)])

    # explode
    _holes = dbu.explode()

    # hull
    hull = gdf.dissolve().convex_hull

    # return holes
    _holes =  _holes[_holes.within(hull.geometry[0])]

    return _holes


def fill_holes(gdf, largest=True, inplace=False):
    """Fill any holes in a geodataframe.

    A hole is also known as a sliver which contains a set of points that:

    - are not contained by any of the geometries in the geoseries
    - are contained in the convex hull of the unary_union of the geoseries

    Parameters
    ----------

    gdf :  GeoDataFrame with polygon (multipolygon) GeoSeries


    largest: boolean (Default: True)
          Merge each hole with its largest (True), or smallest (False) touching neighbor

    inplace: boolean (default: False)
          Change the geoseries of current dataframe


    Returns
    -------

    _holes : GeoDataFrame with hole polygons

    Examples
    --------
    >>> p1 = box(0,0,10,10)
    >>> p2 = Polygon([(10,10), (12,8), (10,6), (12,4), (10,2), (20,5)])
    >>> gdf = geopandas.GeoDataFrame(geometry=[p1,p2])
    >>> gdf.area
    0    100.0
    1     32.0
    dtype: float64
    >>> h = geoplanar.holes(gdf)
    >>> h.area
    array([4., 4.])
    >>> gdf1 = geoplanar.fill_holes(gdf)
    >>> gdf1.area
    0    108.0
    1     32.0
    dtype: float64
    """


    h = holes(gdf)
    if not inplace:
        gdf = gdf.copy()

    for index, row in h.iterrows():
        rdf = geopandas.GeoDataFrame(geometry=[row.geometry])
        neighbors = geopandas.sjoin(left_df=gdf, right_df=rdf, how='inner',
                                    op='intersects')
        if largest:
            left = neighbors[neighbors.area==neighbors.area.max()]
        else:
            left = neighbors[neighbors.area==neighbors.area.min()]
        tmpdf = pandas.concat([left, rdf]).dissolve()
        try:
            geom = tmpdf.loc[0, 'geometry']
            gdf.geometry[left.index] = geopandas.GeoDataFrame(geometry=[geom]).geometry.values
        except:
            print(index, tmpdf.shape)

    return gdf

def missing_interiors(gdf):
    """ Find any missing interiors.

    For a planar enforced polygon layer, there should be no cases of a polygon
    being contained in another polygon. Instead the "contained" polygon is a
    hole in the "containing" polygon.


    Parameters
    ----------

    gdf :  GeoDataFrame with polygon (multipolygon) GeoSeries


    Returns
    -------

    pairs : list
           tuples for each violation (i,j), where i is the index of the
           containing polygon, j is the index of the contained polygon

    Examples
    --------
    >>> p1 = box(0,0,10,10)
    >>> p2 = box(1,1, 3,3)
    >>> p3 = box(7,7, 9,9)
    >>> gdf = geopandas.GeoDataFrame(geometry=[p1,p2,p3])
    >>> mi = geoplanar.missing_interiors(gdf)
    >>> mi
    [(0, 1), (0, 2)]
    """
    contained = gdf.geometry.sindex.query_bulk(gdf.geometry,
                                               predicate="contains")
    pairs = []
    for pair in contained.T:
        i,j = pair
        if i != j:
            pairs.append((i,j))
    return pairs


def add_interiors(gdf, inplace=False):
    """ Add any missing interiors.

    For a planar enforced polygon layer, there should be no cases of a polygon
    being contained in another polygon. Instead the "contained" polygon is a
    hole in the "containing" polygon. This function finds and corrects any such violations.


    Parameters
    ----------

    gdf :  GeoDataFrame with polygon (multipolygon) GeoSeries


    inplace: boolean (default: False)
          Change the geoseries of current dataframe


    Returns
    -------

    gdf : GeoDataFrame


    Examples
    --------
    >>> p1 = box(0,0,10,10)
    >>> p2 = box(1,1, 3,3)
    >>> p3 = box(7,7, 9,9)
    >>> gdf = geopandas.GeoDataFrame(geometry=[p1,p2,p3])
    >>> gdf.area
    0   100.0
    1     4.0
    2     4.0
    >>> mi = geoplanar.missing_interiors(gdf)
    >>> mi
    [(0, 1), (0, 2)]
    >>> gdf1 = geoplanar.add_interiors(gdf)
    >>> gdf1.area
    0    92.0
    1     4.0
    2     4.0
    """
    if not inplace:
        gdf = gdf.copy()

    contained = gdf.geometry.sindex.query_bulk(gdf.geometry,
                                               predicate="contains")
    a, k = contained.shape
    n = gdf.shape[0]
    if k > gdf.shape[0]:
        to_add = contained[:, contained[0] != contained[1]].T
        for add in to_add:
            i, j = add
            gdf.geometry[i] = gdf.geometry[i].difference(gdf.geometry[j])
    return gdf



#!/usr/bin/env python3
import numpy


def trim_overlaps(gdf, largest=True, inplace=False):
    """Trim overlapping polygons


    Parameters
    ----------

    gdf:  geodataframe with polygon geometries

    largest: boolean
             True: trim the larger of the pair of overlapping polygons, False: trim the smaller polygon.

    Returns
    -------

    gdf: geodataframe with corrected geometries

    """
    intersections = gdf.sindex.query_bulk(gdf.geometry, predicate='intersects').T

    if not inplace:
        gdf = gdf.copy()
    if largest:
        for i,j in intersections:
            if i != j:
                left = gdf.geometry[i]
                right = gdf.geometry[j]
                if left.area < right.area:
                    right = gdf.geometry[j].difference(gdf.geometry[i])
                    gdf.geometry[j] = right
                else:
                    left = gdf.geometry[i].difference(gdf.geometry[j])
                    gdf.geometry[i] = left
    else:
        for i,j in intersections:
            if i != j:
                left = gdf.geometry[i]
                right = gdf.geometry[j]
                if left.area > right.area:
                    right = gdf.geometry[j].difference(gdf.geometry[i])
                    gdf.geometry[j] = right
                else:
                    left = gdf.geometry[i].difference(gdf.geometry[j])
                    gdf.geometry[i] = left

    return gdf

def is_overlapping(gdf):
    "Test for overlapping features in geoseries."

    uu = gdf.unary_union
    area_sum = gdf.area.sum()
    if not numpy.isclose(area_sum, uu.area):
        return True
    return False



from distutils.core import setup

import os

# Get __version__ from libpysal/__init__.py without importing the package
# __version__ has to be defined in the first line
with open("geoplanar/__init__.py", "r") as f:
    exec(f.readline())


from setuptools import setup

with open("requirements.txt") as f:
    tests_require = f.readlines()
install_requires = [t.strip() for t in tests_require]

with open("README.md") as f:
    long_description = f.read()



setup(name='geoplanar',
    version=__version__,
    description='Geographic planar enforcement of polygon geoseries',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/sjsrey/geoplanar",
    author='Serge Rey',
    author_email='sjsrey@gmail.com',
    license="3-Clause BSD",
    packages=['geoplanar'],
    package_data={"": ["requirements.txt"]},
    classifiers=[
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: Implementation :: CPython",
    ],
    python_requires=">=3.7",
    install_requires=install_requires,
    zip_safe=False,
)



"""
GeoPlanar

A module for handling planar enforcement for polygon geoseries.
"""

from geoplanar.overlap import trim_overlap, is_overlapping


import geopandas
import libpysal
import numpy
import pandas
from collections import defaultdict
from shapely.geometry import box, LineString, Polygon
from shapely.ops import split

def non_planar_edges(gdf):
    w = libpysal.weights.Queen.from_dataframe(gdf)
    intersections = gdf.sindex.query_bulk(gdf.geometry, predicate='intersects').T
    w1 = defaultdict(list)
    for i,j in intersections:
        if i != j:
            w1[i].append(j)
    missing = defaultdict(set)
    for key in w.neighbors:
        if set(w.neighbors[key]) != set(w1[key]):
            for value in w1[key]:
                pair = [key, value]
                pair.sort()
                l, r = pair
                missing[l] = missing[l].union([r])
    return missing

def holes(gdf):
    """Find holes in a geodataframe"""


    # get box
    b = box(*gdf.total_bounds)

    # form unary union
    u = gdf.unary_union

    # diff of b and u
    dbu = geopandas.GeoDataFrame(geometry=[b.difference(u)])

    # explode
    _holes = dbu.explode()

    # hull
    hull = gdf.dissolve().convex_hull

    # return holes
    _holes =  _holes[_holes.within(hull.geometry[0])]

    return _holes


def fill_holes(gdf):
    h = holes(gdf)
    for index, row in h.iterrows():
        #print(row.geometry.area)
        rdf = geopandas.GeoDataFrame(geometry=[row.geometry])
        neighbors = geopandas.sjoin(left_df=gdf,
                                   right_df=rdf,
                                   how='inner',
                                   op='intersects')
        largest = neighbors[neighbors.area==neighbors.area.max()]
        tmpdf = pandas.concat([largest, rdf]).dissolve()
        gdf.geometry[largest.index] = tmpdf.geometry[0]
        #print(largest.head())
    return gdf

def planar_enforce(gdf):
    uu = gdf.unary_union
    geoms = [uu.intersection(row.geometry) for index, row in gdf.iterrows()]
    return geopandas.GeoDataFrame(geometry=geoms)

def is_planar_enforced(gdf):
    if is_overlapping(gdf):
        return False
    if non_planar_edges(gdf):
        return False
    _holes = holes(gdf)
    if _holes.shape[0] > 0:
        return False
    return True

def check_validity(gdf):
    results = {}

    _edges = non_planar_edges(gdf)
    if _edges:
        results['edges'] = _edges

    return results

def fix_npe_edges(gdf):
    "Fix all npe intersecting edges in geoseries"
    edges = non_planar_edges(gdf)
    print(edges)
    for key in edges:
        poly_a = gdf.iloc[key].geometry
        for j in edges[key]:
            poly_b = gdf.iloc[j].geometry
            new_a, new_b = insert_intersections(poly_a, poly_b)
            poly_a = new_a
            gdf.geometry[key] = new_a
            gdf.geometry[j] = new_b
    return gdf

def insert_intersections(poly_a, poly_b):
    """Correct two npe intersecting polygons by inserting intersection points on intersecting edges
    """
    lin_a = LineString(list(poly_a.exterior.coords))
    lin_b = LineString(list(poly_b.exterior.coords))
    new_polys = []
    pts = lin_a.intersection(lin_b)
    for lin in [lin_a, lin_b]:
        splits = split(lin, pts)
        coords = [list(zip(*line.coords.xy)) for line in list(splits)]
        new_poly = coords[0]
        for coord in coords[1:]:
            if coord[0] == new_poly[-1]:
                new_poly.extend(coord[1:])
            else:
                new_poly.extend(coord)
        new_polys.append(Polygon(new_poly))
    return new_polys
